/*  MIDIreceiverHS_Rnn.ino
  this script by Holger Suchi receive MIDI events on MIDIin and actuates solenoids
  - modificated for an Akkordzither Hopf 5/21 by Sören Kassel
    MIDIreceiverHS_ZitherAutoAkkord.ino
  - last action 17.02.2021
  
  harware used:
  - Adrduino UNO
  - MIDI shield
    connected NOT to RX/TX but to 2/3 using software serial
    to have hardware serial available for debugging messages
  - Portexpander Board by Holger Suchi
  
  todos:
  - implement I2C 32 bit expander --- ok
  - implement multiple portexpanders at the I2C
  - implement auto note off for e.g. Glockenspiel --- ok (DI5)
  - implement switch  debug on/off --- open
  - implement switch auto_on_mode on/off --- ok (DI6)
  - implement poti to change note_length extern --- ok (DI8)
  - implement bluetooth receiver --- open
*/

#define debug 1 // 1 enables debugging messages via HW serial @57600baud

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
//---    global settings & declarations & hardware definitions               ---
//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
#include <SoftwareSerial.h>
SoftwareSerial MIDIserial(2, 3); // RX, TX
#define MIDI_RATE 31250 // to be used at serial
#define MIDIrxPin 2
#define MIDItxPin 3

#include <Wire.h>
#include "Adafruit_MCP23017.h"
Adafruit_MCP23017 mcp1;
Adafruit_MCP23017 mcp2;

//------------------------------------------------------------------------------
//---   NeopixelSetup                                                        ---
//------------------------------------------------------------------------------
#include <Adafruit_NeoPixel.h>
#define PIN 7
#define NUMPIXELS 33
Adafruit_NeoPixel pixels(NUMPIXELS, PIN, NEO_GRB + NEO_KHZ800);


//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
//---    global types & variables & constantes                               ---
//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
#define ON HIGH
#define OFF LOW
typedef struct {  // holding incoming MIDI message info
  byte raw;       // incoming byte read from MIDI serial
  byte cmd;       // command+channel, first message byte
  byte chan;      // lower nibbe of cmd
  byte note;      // first data byte, holding note#
  byte vel;       // scond data byte, holding velocity (0==off)
} event_t;

event_t ev;

enum MidiRecState { MRS_waitforfirst, MRS_getnote, MRS_getvel };
static MidiRecState MRS = MRS_waitforfirst;

//variables setup
byte poex_address = 0;     // I2C address sent to port expander;
// as 2x MCP23017 used in my 32bit port expander, two consequtive adresses used

byte poex_address0 = 0;    // I2C address of pipe controller
// for future extensions with more then one port expander
//byte poex_address1 = 0;    // I2C address of second controller
//byte poex_address2 = 0;    // I2C address of third controller
//byte poex_address3 = 0;    // I2C address of fourth controller

unsigned long poex_data = 0;  // 4 byte / 32bit sent to port expander
//each bit correponds to one output
unsigned long notemask = 0;   // 4 byte, each bit correponds to one output

int pipe = 0;
byte pn; // pipe number  (1..21, 21..24 for testing and intonation valves)
//******************************************************************************
// assignement of expander output per midi note number (127 midi notes)
// return zero for invalid note

// Zither String 1-21 Melodiesaiten: C3, C#3, D3, E3, F3, F#3, G3, G#3, A3, B3, C4, C#4, D4, E4, F4, F#4, G4 G#4, A4, B4, C5
// Zither String 22 ,24, 26, 28, 30 Grundsaite Akkorde: A1, D1, F1, G1, C2
// Zither String 23, 25, 27, 29, 31 je 3 Akkordsaiten zusammengefasst: E4, A3, C4; D4, A3, F#1; F4, C3, A3; D3, B2, G2; E3, C3, G2
// Zither String 23, 25, 27, 29, 31 sollen später zusammen mit 22, 24, 26, 28, 30 zeitverzögert anschlagen (pipe_on_delay) --- ok
// Anordnung der pn in midimap[] muss noch an Verdrahtung angepasst werden --- ok

byte midimap[] = {\
                  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, \
                  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, \
                  0, 0, 0, 0, 0, 0, 24, 0, 0, 26, 0, 28, 0, 22, 0, 0, \
                  30, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 2, 3, 0, \
                  4, 5, 6, 7, 8, 9, 0, 10, 11, 12, 13, 0, 14, 15, 16, 17, \
                  18, 19, 0, 20, 21, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, \
                  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, \
                  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
                 };

// array storing the auto note-off time in millis for each valid note
// array ranges from 0 to 31 to match pipe number; element[0] unused!
// compares to millis(), so FFFFFFFF is always larger then millis()
unsigned long pipe_off_time[] = {\
                                 0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF, \
                                 0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF, \
                                 0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF, \
                                 0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF, \
                                 0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF, \
                                 0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF, \
                                 0xFFFFFFFF, 0xFFFFFFFF
                                };
// array storing the auto note-on time in millis for each valid note
// array ranges from 0 to 31 to match pipe number; element[0] unused!
// compares to millis(), so FFFFFFFF is always larger then millis()
unsigned long pipe_on_time[] = {\
                                0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF, \
                                0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF, \
                                0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF, \
                                0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF, \
                                0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF, \
                                0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF, \
                                0xFFFFFFFF, 0xFFFFFFFF
                               };
//array define the pipe_on_delay time in ms for each valid note
unsigned int pipe_on_delay[] = {\
                                0, 0, 0, 0, 0, 0, 0, 0, 0, 0, \
                                0, 0, 0, 0, 0, 0, 0, 0, 0, 0, \
                                0, 0, 0, 20, 0, 20, 0, 20, \
                                0, 20, 0, 20
                               };


// array define the note_length time in ms for each valid note
byte note_length[] = {\
                      0, 5, 5, 5, 5, 5, 5, 5, 5, 5, \
                      5, 5, 5, 5, 5, 5, 5, 5, 5, 5, \
                      5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5
                     };
                     
byte led_nr[] = {\
                       0, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, \
                      22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, \
                       1, 1, 3, 3, 5, 5, 8, 8, 10, 10, 
                     };
                     
//const byte auto_off_mode = 1;
int auto_off_mode = 1;
// 1 = active auto off; 0 = no auto off
//const byte auto_on_mode = 1;
int auto_on_mode = 1;
// 1 = active auto on; 0 = no auto on
unsigned long tim_off;    // marker for auto-off time
unsigned long tim_on;    // marker for auto-on time

// time to reset a note in auto_off mode; actualy replaced with note_length[]; later used with poti extern
unsigned int play_time = 5;

//set input A3 as poti for play_time
int analogPin = A3;
int pot = 0;
int ext = 0; //1=Poti, 0=Array, (DI8)



//******************************************************************************

// live-blink LED connected to output 32
unsigned long time2blink; // time marker for liveblinker

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
void setup()
{
#if debug
  Serial.begin(57600);
  while (!Serial) {
    ; // wait for serial port to connect. Needed for Native USB only
  }
  Serial.println("Serial initialized");
#endif

  // set up the SoftwareSerial port
  // define pin modes for tx, rx:
  pinMode(MIDIrxPin, INPUT);
  pinMode(MIDItxPin, OUTPUT);
  // set the data rate for the SoftwareSerial port
  MIDIserial.begin(MIDI_RATE);

#if debug
  Serial.println("MIDIserial initialized");
#endif

  // initialize I2C bus communication to 2x MCP23017
  Wire.setClock(400000); // set fast mode I2C speed (does not matter actually)
  mcp1.begin(0);     // use address 0   !! future more...
  mcp2.begin(1);     // use address 1   !! future more...
  // initialize expander ports

  for (int n = 0; n < 16; n++) {
    mcp1.pinMode(n, OUTPUT);
  }

  
  for (int n = 0; n < 16; n++) {
    mcp2.pinMode(n, OUTPUT);
  }
  
  //poex_data = 0xFFFFFFFF; // all note on
  poex_data = 0x00000000; // all note off
  // initialize port expander here
  mcp1.writeGPIOAB((uint16_t)(poex_data & 0x0000FFFF));
  mcp2.writeGPIOAB((uint16_t)(poex_data >> 16));

#if debug
  Serial.println("Alle Relais Aus");
#endif

  // time to first liveblink
  time2blink = millis() + 1000;

 pixels.begin(); //INITIALIZE NeoPixel strip object
 pixels.clear(); 
  
  pinMode(5, INPUT_PULLUP);
  pinMode(6, INPUT_PULLUP);
  pinMode(8, INPUT_PULLUP);
  
} // setup
//------------------------------------------------------------------------------
//------------------------------------------------------------------------------


//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
void loop()
{
  
  //----------------------------------------------------------------------------
  //state machine to look for MIDIserial data
  if (MIDIserial.available()) {
    // read the incoming byte:
    
    ev.raw = MIDIserial.read(); // read one byte
    switch (MRS) {              // MRS - midi receive status
      case MRS_waitforfirst:
        // look for a status-byte, our channel, note on/off
        // !! future: distinguish channels for different expanders
        // serve only 0x9n or 0x8n
        if (((ev.raw & 0xF0) == 0x90) || ((ev.raw & 0xF0) == 0x80)) 
        {
          ev.cmd = ev.raw;
          ev.chan = ev.raw & 0x0F ;
          MRS = MRS_getnote;
        }
        break;
      case MRS_getnote:
        // get the midi note to play or stop
        if (!(ev.raw & 0x80)) {   // midi notes ranges only 0..127
          ev.note = ev.raw;
          MRS = MRS_getvel;
        }
        else {                    // invalid note byte
          MRS = MRS_waitforfirst; // reset state machine
        }
        break;
      case MRS_getvel:
        // get the velocity
        if (!(ev.raw & 0x80)) {   // midi velocity ranges only 0..127
          ev.vel = ev.raw;
          // event service here
          
          playNote();             // play a note
          
        }
        MRS = MRS_waitforfirst;   // reset state machine to start
    } // switch MRS
  } // if (MIDIserial.available())
  //----------------------------------------------------------------------------

  //----------------------------------------------------------------------------
  // handle auto_off mode here
  if (auto_off_mode) //1 = Glockenspiel(Standard); Notenlänge aus Array
                     //0 = 
  {
    for (pipe = 1; pipe < 32; pipe++) {
      // actual time larger then intended off-time?
      if (millis() > pipe_off_time[pipe]) {
        // then set this output off
        notemask = 0;
        // determine bit position in poex_data from note number
        notemask = (unsigned long) 1 << (pipe - 1); 
        // corresponding bit clear
        poex_data &= ~notemask; 
        mcp1.writeGPIOAB((uint16_t)(poex_data & 0x0000FFFF));
        mcp2.writeGPIOAB((uint16_t)(poex_data >> 16));

        
pixels.setPixelColor(led_nr[pn], pixels.Color(0, 0, 0)); 
pixels.show();
   
        // reset served element in pipe_off_time array
        pipe_off_time[pipe] = 0xFFFFFFFF;
        
#if debug
        Serial.print(" auto_off pipe: ");
        Serial.println(pipe);
        
        
#endif
      }//if (millis() > pipe_off_time
    }//for (pipe = 1; pipe < 32
  }//if (auto_off_mode)
  //----------------------------------------------------------------------------
  //----------------------------------------------------------------------------
  // handle auto_on mode here
  if (!(auto_on_mode)) //1 = one Note Akkord(Standard); 0 = Full Akkord delayed from Array "pipe_on_delay"
  {
    for (pipe = 22; pipe < 32; pipe++) {  // not necessary to test all pipes here? 
      // actual time larger then intended on-time?
      if (millis() > pipe_on_time[pipe]) {
        // then set this output on
        notemask = 0;
        // determine bit position in poex_data from note number
        notemask = (unsigned long) 1 << (pipe - 1); 
        
        // corresponding bit set
        poex_data |= notemask;
        mcp1.writeGPIOAB((uint16_t)(poex_data & 0x0000FFFF));
        mcp2.writeGPIOAB((uint16_t)(poex_data >> 16));

        // reset served element in pipe_off_time array
        tim_off = millis() + note_length[pipe]; //get activation time from array and add offset
        
        // store in pipe_off_time array
        pipe_off_time[pipe] = tim_off;
        pipe_on_time[pipe] = 0xFFFFFFFF;
        
#if debug
        Serial.print(" auto_on pipe: ");
        Serial.println(pipe);
#endif
      }//if (millis() > pipe_on_time
    }//for (pipe = 22; pipe < 32
  }//if (auto_on_mode)
  //----------------------------------------------------------------------------

  //----------------------------------------------------------------------------
  // life_blinker
  // I have a LED connected to output 32
  if ( millis() > time2blink) {
    time2blink = millis() + 100;
    poex_data ^= 0x80000000;  // exor Output 32
    mcp2.writeGPIOAB((uint16_t)(poex_data >> 16));
#if debug
    Serial.println("...warte auf MIDI-Daten...");
    pixels.clear();
    pixels.show();
#endif

  

  }// if ( millis() > time2blink)
  //----------------------------------------------------------------------------
  ext = digitalRead(8);
  pot = analogRead(analogPin)/50; //analog in 0-20
  play_time = pot;
  auto_on_mode = digitalRead(6);
  auto_off_mode = digitalRead(5);

  
  

} // loop
//----------------------------------------------------------------------------
//----------------------------------------------------------------------------

// event service routine

void playNote(void) {
  pn = midimap[ev.note];    // get assigned output

  if ( pn ) { // on valid notes only; pn <> 0
    notemask = 0;
    // determine bit position in poex_data from note number
    notemask = (unsigned long) 1 << (pn - 1); // corresponding bit set

    // !! future: distingish between different midi channels...
    // possibly for all channels?
    // poex_address = poex_address0;
    if (ev.chan = 0) poex_address = poex_address0;
    // if (ev.chan = 1) poex_address = poex_address1;
    // if (ev.chan = 2) poex_address = poex_address2;
    // if (ev.chan = 3) poex_address = poex_address3;

    // note ON or OFF
    if ((ev.cmd & 0x10) && ev.vel) { // note on and vel > 0 ?
      
      // noteON
      poex_data |= notemask;
      // on each note being activated set the marker for auto_off
  if (digitalRead(8) == 0)
  {  
      Serial.print("Wert von Poti ");
      Serial.println(play_time);
      tim_off = millis() + play_time; //get activation time from extern poti
  }
  else
  {

  
      Serial.print("Wert von Array ");
      Serial.println(note_length[pn]);
      
      tim_off = millis() + note_length[pn]; //get activation time from array and add offset
  }   
      pipe_off_time[pn] = tim_off;    // store in pipe_off_time array
     


      // handle the 5 auto_on notes here
    switch (pn) {
      case 22:
        tim_on = tim_off - note_length[pn] + pipe_on_delay[pn + 1];
              // tim_off war der Zeitpunkt des Anschlags der 22;
              // subtract the offset
              // add the delay until on
              // pn+1 if next is concerned, or fix value here

        pipe_on_time[pn + 1] = tim_on;  // store in pipe_on_time array for related pipe
        break;
      case 24:
        tim_on = tim_off - note_length[pn] + pipe_on_delay[pn + 1]; // tim_off war der Zeitpunkt des Anschlags der 24;
        pipe_on_time[pn + 1] = tim_on;  // store in pipe_on_time array for related pipe
        break;
      case 26:
        tim_on = tim_off - note_length[pn] + pipe_on_delay[pn + 1]; // tim_off war der Zeitpunkt des Anschlags der 26;
        pipe_on_time[pn + 1] = tim_on;  // store in pipe_on_time array for related pipe
        break;
      case 28:
        tim_on = tim_off - note_length[pn] + pipe_on_delay[pn + 1]; // tim_off war der Zeitpunkt des Anschlags der 28;
        pipe_on_time[pn + 1] = tim_on;  // store in pipe_on_time array for related pipe
        break;
      case 30:
        tim_on = tim_off - note_length[pn] + pipe_on_delay[pn + 1]; // tim_off war der Zeitpunkt des Anschlags der 30;
        pipe_on_time[pn + 1] = tim_on;  // store in pipe_on_time array for related pipe
        break;
    }// switch (pn)
    
    } else {  // note off or vel = 0
      
      // noteOFF
      poex_data &= ~notemask;
      pipe_off_time[pn] = 0xFFFFFFFF;    // reset pipe_off_time array
      
     }
    //main output function here
    mcp1.writeGPIOAB((uint16_t)(poex_data & 0x0000FFFF));
    mcp2.writeGPIOAB((uint16_t)(poex_data >> 16));

pixels.setPixelColor(led_nr[pn], pixels.Color(12, 12, 12)); 
pixels.show();

#if debug
    // debug info
    Serial.print(" Aktiviere Relais: ");
    Serial.println(pn);
    
#endif
  }// on valid notes only


#if debug
  // debug info
  Serial.print("midi-cmd: ");
  Serial.print(ev.cmd, DEC);
  Serial.print(" midi-note: ");
  Serial.print(ev.note, DEC);
  Serial.print(" vel: ");
  Serial.println(ev.vel, DEC);
#endif

}//void playNote(void)
